{-# LANGUAGE FlexibleContexts #-}
{-# OPTIONS_GHC -Wall         #-}

module HarmTrace.Audio.Annotations where

-- import Text.ParserCombinators.UU hiding (join)
-- import Text.ParserCombinators.UU.BasicInstances
-- import Data.ListLike.Base (ListLike)

import Data.Maybe (isJust,fromJust)

import HarmTrace.Audio.Parser
import HarmTrace.Audio.ChordTypes
import HarmTrace.Base.MusicRep
import HarmTrace.Base.Parsing
import HarmTrace.Tokenizer.Tokenizer (parseDegrees, parseDegree)

-- perhaps this file should be moved to the tokeniser module, because it is 
-- is very related to tokenising

--------------------------------------------------------------------------------
-- Harmonically analysing Chord Annotations
--------------------------------------------------------------------------------

preProcess  :: [ChordSegment] -> [ChordLabel]
preProcess = preProcess' 0 
preProcess' :: Int -> [ChordSegment] -> [ChordLabel]
preProcess' _ [] = []
preProcess' ix ((TimedData c@(Chord r sh ad _loc dur) _on _off):ns) 
  | isNoneChord c = preProcess' (ix) ns
  | otherwise     = (Chord r sh ad ix dur) : preProcess' (ix+1) ns

--------------------------------------------------------------------------------
-- Chords
--------------------------------------------------------------------------------

parseAnnotationData :: Parser ChordAnnotation
parseAnnotationData = pListSep_ng pLineEnd pChordSegment <* pLineEnd

pChordSegment :: Parser ChordSegment
pChordSegment = timedData' <$> pNumData <* pSpaceTab
                           <*> pNumData <* pSpaceTab <*> pChord


pChord :: Parser ChordLabel
pChord = f <$> parseRoot <*> pMaybe (pSym ':' *> parseShorthand)
                         <*> (parseDegrees `opt` []) 
                         -- there might be a basenote inversion annotation
                         -- which we currently ignore
                         <*> pMaybe (pSym '/' *> parseDegree) where
           -- if there is no root note, there is no chord type
           -- Chord root shorthand additions start duration
  f r sh _ _ | r == (Note Nothing N) = Chord r None [] 0 1 
             | isJust sh             = Chord r (fromJust sh) [] 0 1 
             | otherwise             = Chord r Maj  [] 0 1 

--------------------------------------------------------------------------------
-- Keys
--------------------------------------------------------------------------------

parseKeyAnnotationData :: Parser KeyAnnotation
parseKeyAnnotationData = pListSep_ng pLineEnd pKeySegment <* pLineEnd

pKeySegment :: Parser KeySegment
pKeySegment = timedData'  <$> pNumData <*  pSpaceTab
                          <*> pNumData <*  pSpaceTab 
                          <*> (pKey    <|> pKeyNone)

pKey :: Parser Key
pKey = Key <$ pString "Key" <* pSpaceTab <*> parseRoot <*> pMode

pKeyNone :: Parser Key
pKeyNone = (Key (Note Nothing N) MajMode) <$ pString "Silence"

pMode :: Parser Mode
pMode  =     MajMode <$ pString ""
         <|> MinMode <$ pString ":minor"
         <|> MinMode <$ pString ":aeolian"
         <|> MajMode <$ pString ":major"
         <|> MinMode <$ pString ":dorian"
         <|> MajMode <$ pString ":mixolydian" -- this must be solved differently
         <|> MajMode <$ pString ":modal"      -- and this too....
         

--------------------------------------------------------------------------------
-- General Parsers and Utils
--------------------------------------------------------------------------------

timedData' :: NumData -> NumData -> a -> TimedData a
timedData' on off chrd = TimedData chrd on off

pSpaceTab :: Parser Char
pSpaceTab =  pSym ' ' <|> pSym '\t'

-- begin stealing from Tokenizer.hs (added/changed some ad hoc instances)

parseShorthand :: Parser Shorthand
parseShorthand =     Maj      <$ pString "maj"
                 <|> Min      <$ pString "min"
                 <|> Dim      <$ pString "dim"
                 <|> Aug      <$ pString "aug"
                 <|> Maj7     <$ pString "maj7"
                 <|> Min7     <$ pString "min7"
                 <|> Sev      <$ pString "7"
                 <|> Dim7     <$ pString "dim7"
                 <|> HDim7    <$ pString "hdim" <* opt (pSym '7') '7'
                 <|> MinMaj7  <$ pString "minmaj7"
                 <|> Maj6     <$ pString "maj6"
                 <|> Maj6     <$ pString "6"
                 <|> Min6     <$ pString "min6"
                 <|> Nin      <$ pString "9"
                 <|> Maj9     <$ pString "maj9"
                 <|> Min9     <$ pString "min9"
                 <|> Maj      <$ pString ""-- this is not really nice, but works
                 <|> Sus4     <$ pString "sus4"
                 <|> Sus2     <$ pString "sus2"
                 <?> "Shorthand"

parseRoot :: Parser Root
parseRoot =     Note Nothing   A  <$ pSym 'A'
            <|> Note Nothing   B  <$ pSym 'B'
            <|> Note Nothing   C  <$ pSym 'C'
            <|> Note Nothing   D  <$ pSym 'D'
            <|> Note Nothing   E  <$ pSym 'E'
            <|> Note Nothing   F  <$ pSym 'F'
            <|> Note Nothing   G  <$ pSym 'G'
            <|> Note Nothing   N  <$ pSym 'N' -- for no chord
            <|> Note (Just Fl) A <$ pString "Ab"
            <|> Note (Just Fl) B <$ pString "Bb"
            <|> Note (Just Fl) C <$ pString "Cb"
            <|> Note (Just Fl) D <$ pString "Db"
            <|> Note (Just Fl) E <$ pString "Eb"
            <|> Note (Just Fl) F <$ pString "Fb"
            <|> Note (Just Fl) G <$ pString "Gb"
            <|> Note (Just Sh) A <$ pString "A#"
            <|> Note (Just Sh) B <$ pString "B#"
            <|> Note (Just Sh) C <$ pString "C#"
            <|> Note (Just Sh) D <$ pString "D#"
            <|> Note (Just Sh) E <$ pString "E#"
            <|> Note (Just Sh) F <$ pString "F#"
            <|> Note (Just Sh) G <$ pString "G#" <?> "Chord root"

-- end stealing from Tokenizer.hs
